unit TreeViewer;
  { To view (possibly malformed) binary trees of vectors }
  { Tom Verhoeff (TUE, 2007) }

interface

uses
  Vectors, PointSets;

const
  OutFileName: String = 'TreeViewer.out'; // file name can be changed

procedure WriteTree(const msg, name: String; t: PNode);
  // pre: true (t does not have to be a TBinTree or TBST)
  // post: timestamp, msg, name, t, and t^ written to output file

implementation

uses
  SysUtils, Classes;

procedure WriteTree(const msg, name: String; t: PNode);

  const
    IndentStep = '  '; // to indent one more level;

  var
    OutFile: TextFile; // handle to output file
    Visited: TList; // list of non-nil pointers seen so far
    CollisionCount: Integer; // number of collisions
    PrecVector: TVector; // preceding TVector if any else nil
    NonIncCount: Integer; // number of non-increasing neighboring pairs

  procedure WriteTreeRec(t: PNode; const indent: String);
      // pre: true
      // post: write t and, if not yet visited, t^
      //   uses in-order to simplify check on order
    begin
      if t = nil
      then begin
        Writeln(OutFile, 'nil');
      end
      else { t <> nil } begin
        Write(OutFile, IntToHex(Integer(t), 8));
        if Visited.IndexOf(t) = -1
        then { not yet visited } begin
          Writeln(OutFile, ' pointing at TNode with');
          Visited.Add(t);
          with t^ do begin
            Write  (OutFile, indent, 'FLeft   = ');
            WriteTreeRec(FLeft, indent + IndentStep);

            Write(OutFile, indent, 'FVector = ', FVector.ToStr);
            if (PrecVector = nil) or (PrecVector.Compare(FVector) = -1)
            then { order is OK } begin
              Writeln(OutFile);
            end
            else { (PrecVector <> nil) and (PrecVector < FVector) } begin
              Writeln(OutFile, ' (out of order w.r.t. preceding TNode)');
              Inc(NonIncCount);
            end;
            PrecVector := FVector;

            Write  (OutFile, indent, 'FRight  = ');
            WriteTreeRec(FRight, indent + IndentStep);
          end; { with }
        end
        else { already visited } begin
          Writeln(OutFile, ' (see above)');
          Inc(CollisionCount);
        end;
      end;
    end;

  begin
    Assign(OutFile, OutFileName);
    if FileExists(OutFileName) then begin
      Append(OutFile);
    end
    else { file does not yet exist } begin
      Rewrite(OutFile);
    end;
    Writeln(OutFile, DateTimeToStr(Now), ' # ', msg); // time stamp and message
    Visited := TList.Create; // empty list
    CollisionCount := 0;
    PrecVector := nil;
    NonIncCount := 0;
    Write(OutFile, name, ' = ');

    WriteTreeRec(t, IndentStep);

    Writeln(OutFile, Visited.Count, ' nodes written');
    if CollisionCount <> 0 then begin
      Writeln(OutFile, 'WARNING: ', name,
        ' is not a valid TBinTree, having ', CollisionCount, ' violations');
    end; { if }
    if NonIncCount <> 0 then begin
      Writeln(OutFile, 'WARNING: ', name,
        ' is not a valid TBST, having ', NonIncCount, ' order violations');
    end; { if }
    Writeln(OutFile);
    Visited.Free;
    CloseFile(OutFile);
  end;

end.
