unit Vectors;
  { ADT for vectors in Real^2 }
  { Tom Verhoeff (TUE, 2007) }

interface

uses
  Graphics;

type
  TVector = class(TObject) // Abstract Data Type
  private
    FX: Real; // X coordinate
    FY: Real; // Y coordinate

  public
    // construction -------------------
    constructor Create(AX, AY: Real);
      // pre: true
      // post: X = AX /\ Y = AY

    // basic queries ------------------
    property X: Real read FX;
    property Y: Real read FY;

    // derived queries ----------------
    function IntX: Integer;
      // pre: true
      // ret: Round(X)
    function IntY: Integer;
      // pre: true
      // ret: Round(Y)
    function ToStr: String;
      // pre: true
      // ret: vector as string
    function Compare(AVector: TVector): Integer;
      // pre: true
      // ret: -1 if Self < AVector, 0 if Self = AVector, else 1
      //   lexicographic order on pairs

    procedure Draw(ACanvas: TCanvas; DX, DY, CX, CY, Angle: Integer);
      // pre: true
      // effect: Self drawn on ACanvas, rotated around CX, CY over Angle,
      //         then shifted over DX, DY
  end;

implementation

uses
  SysUtils, Math;

constructor TVector.Create(AX, AY: Real);
begin
  inherited Create;
  FX := AX;
  FY := AY;
end;

function TVector.IntX: Integer;
begin
  Result := Round(X);
end;

function TVector.IntY: Integer;
begin
  Result := Round(Y);
end;

procedure TVector.Draw(ACanvas: TCanvas; DX, DY, CX, CY, Angle: Integer);
var
  VX, VY: Integer; // transformation of Self
  AngleRad: Real; // Angle in radians
  S, C: Real; // sine and cosine of Angle
begin
  AngleRad := DegToRad(Angle);
  S := sin(AngleRad);
  C := cos(AngleRad);
  // matrix for rotation over Angle
  //   [ C  -S ]
  //   [ S   C ]
  VX := Round(C * (X-CX) - S * (Y-CY) + CX + DX);
  VY := Round(S * (X-CX) + C * (Y-CY) + CY + DY);
//  if IsConsole then begin // show testing output
//    Writeln(Format(
//      'AngleRad, S, C, VX, VY = %1.2F, %1.2F, %1.2F, %D, %D',
//      [AngleRad, S, C, VX, VY]));
//  end; { if }
  with ACanvas do begin
    MoveTo(VX, VY);
    // draw a filled circle with diameter 2 * Pen.Width
    Ellipse(VX-Pen.Width, VY-Pen.Width, VX+Pen.Width, VY+Pen.Width);
  end; { with }
end;

function TVector.ToStr: String;
begin
  Result := Format('(%1.2F, %1.2F)', [X, Y]);
end;

function TVector.Compare(AVector: TVector): Integer;
begin
  if X < AVector.X then Result := -1
  else if X > AVector.X then Result := 1
  else { X = AVector.X } begin
    if Y < AVector.Y then Result := -1
    else if Y > AVector.Y then Result := 1
    else { Y = AVector.Y } Result := 0
  end;
end;

end.
