program ProcType;
  { Illustrates Binary (Search) Trees with pointers }

uses
  Math; { for Max }

type
  TData = Integer; // or any other type definition
  PNode = ^TNode;
  TNode = // node of binary tree over TData
  record
    FData: TData; // value stored in this node
    FLeft: PNode; // left subtree
    FRight: PNode; // right subtree
  end;
  TBTData = PNode; // Binary Trees over TData
    // data invariants
    //   I0: all root paths via FLeft/FRight terminate in nil
    //   I1: if two root paths both end in u then u = nil
    // Abstraction function for t: TBTData
    // empty tree if t=nil else <t^.FLeft, t^.FData, t^.FRight>

function MakeBTData(a: TData; L, R: TBTData): TBTData;
  // pre: L and R do not share any nodes
  // ret: <L, a, R>
  begin
    New(Result);
    with Result^ do begin
      FData := a;
      FLeft := L;
      FRight := R
    end; { with }
    // Result satisfies I0 and I1
  end;

function Height(t: TBTData): Integer;
  // pre: true
  // ret: h.t
  begin
    if t = nil then begin
      Result := 0;
    end
    else { t <> nil } begin
      Result := 1 + Max ( Height(t^.FLeft), Height(t^.FRight) );
      // termination guaranteed because of I0
    end;
  end;

procedure PreOrder(t: TBTData);
  // pre: true
  // post: all nodes of t have been processed in pre-order
  begin
    if t = nil then begin
      // skip
    end
    else { t <> nil } begin
      with t^ do begin
        Write(' ', FData);
        PreOrder(FLeft);
        PreOrder(FRight);
      end; { with }
    end;
  end;

procedure InOrder(t: TBTData);
  // pre: true
  // post: all nodes of t have been processed in in-order
  begin
    // Slightly optimized code
    if t <> nil then begin
      with t^ do begin
        InOrder(FLeft);
        Write(' ', FData);
        InOrder(FRight);
      end; { with }
    end; { if }
  end;

type
  TDataAction = procedure ( const AData: TData );

procedure PostOrder(t: TBTData; ADataAction: TDataAction);
  // pre: true
  // post: all nodes of t have been processed in post-order
  //       by applying ADataAction
  begin
    if t <> nil then begin
      with t^ do begin
        PostOrder(FLeft, ADataAction);
        PostOrder(FRight, ADataAction);
        ADataAction(FData);
      end; { with }
    end; { if }
  end;

procedure WriteData(const AData: TData);
  begin
    Write(' ', AData);
  end;

var
  count: Integer; // global counter for CountData

procedure CountData(const AData: TData {; globvar count});
  // effect: count := count + 1
  begin
    count := count + 1
  end;

procedure DisposeBTData(var t: TBTDAta);
begin
  if t <> nil then begin
    with t^ do begin
      DisposeBTData(FLeft);
      DisposeBTData(FRight);
      // if necessary, dispose FDAta
    end; { with }
    Dispose(t);
  end; { if }
end;

type
  TBSTData = TBTData; // Binary Search Tree over TData
    // data invariant
    //   I2: in-order list is strictly increasing
    // for every subtree <L, a, R>: data in L < a < data in R
  TCompareData = ( ls, eq, gt ); // less, equal, greater

function CompareData(X, Y: TData): TCompareData;
begin
  if X < Y then Result := ls
  else if X = Y then Result := eq
  else { X > Y } Result := gt
end;

function FindRec(const X: TData; t: TBSTData): PNode;
  // pre: true
  // ret: r with t -*-> r /\ r^.FData = X if X in t else nil
  begin
    if t = nil then Result := nil
    else { t <> nil } begin
      with t^ do begin
        case CompareData(X, FData) of
          ls: Result := FindRec(X, FLeft);
          eq: Result := t;
          gt: Result := FindRec(X, FRight);
        end; { case }
      end; { with }
    end; { else }
  end;

function FindIter(const X: TData; t: TBSTData): PNode;
  // pre: t = T
  // ret: r with t -*-> r /\ r^.FData = X if X in T else nil
  begin
    Result := nil;

    // inv: T -*-> t -*-> Result  /\  X in T == X in t /\ 
    //   (Result <> nil ==> Result^.FData = X)
    while t <> Result do begin { t <> nil }
      with t^ do begin
        case CompareData(X, FData) of
          ls: t :=  FLeft;
          eq: Result := t;
          gt: t := FRight;
        end; { case }
      end; { with }
    end; { while }

  end;

procedure Insert(const X: TData; var t: TBSTData);
  // pre: true
  // post: t = initial t extended with X
  begin
    if t = nil then t := MakeBTData(X, nil, nil)
    else { t <> nil } begin
      with t^ do begin
        case CompareData(X, FData) of
          ls: Insert(X, FLeft);
          eq: { skip }; // already present, do not duplicate
          gt: Insert(X, FRight);
        end; { case }
      end; { with }
    end; { else }
  end;

var
  t: TBTData; // a binary tree over TData
  bst: TBSTData; // a binary search tree over TData
  p: PNode; // to capture result of Find

begin
  Writeln('Binary Tree Examples');
  t := MakeBTData(7,
         MakeBTData(3,
           MakeBTData(2, nil, nil),
           MakeBTData(5, nil, nil)),
         MakeBTData(11, nil, nil));

  Writeln('Height = ', Height(t));

  Write('Pre-order:');
  PreOrder(t);
  Writeln;

  Write('In-order:');
  InOrder(t);
  Writeln;

  Write('Post-order:');
  // WriteData has type TDataAction
  PostOrder(t, WriteData);
  Writeln;

  count := 0;
  PostOrder(t, CountData);
  Writeln('Count = ', count);

  DisposeBTData(t);

  Writeln('Binary Search Tree Examples');
  bst := nil;
  Insert(7, bst);
  Insert(3, bst);
  Insert(2, bst);
  Insert(11, bst);
  Write('In-order:');
  InOrder(bst);
  Writeln;

  p := FindRec(2, bst);
  if p = nil then Writeln('2 not found')
  else Writeln('2 found');
  p := FindIter(5, bst);
  if p = nil then Insert(5, bst)
  else Writeln('5 found');
  Write('In-order:');
  InOrder(bst);
  Writeln;
  
  DisposeBTData(bst);
end.
